import tempfile
import time

import os
import os.path


from ovos_utils.log import LOG


def get_ipc_directory(domain=None, config=None):
    """Get the directory used for Inter Process Communication

    Files in this folder can be accessed by different processes on the
    machine.  Useful for communication.  This is often a small RAM disk.

    Args:
        domain (str): The IPC domain.  Basically a subdirectory to prevent
            overlapping signal filenames.
        config (dict): mycroft.conf, to read ipc directory from

    Returns:
        str: a path to the IPC directory
    """
    if config is None:
        try:
            from ovos_config.config import read_mycroft_config
            config = read_mycroft_config()
        except ImportError:
            LOG.warning("Config not provided and ovos_config not available")
            config = dict()
    path = config.get("ipc_path")
    if not path:
        # If not defined, use /tmp/mycroft/ipc
        path = os.path.join(tempfile.gettempdir(), "mycroft", "ipc")
    return ensure_directory_exists(path, domain)


def ensure_directory_exists(directory, domain=None):
    """ Create a directory and give access rights to all

    Args:
        domain (str): The IPC domain.  Basically a subdirectory to prevent
            overlapping signal filenames.

    Returns:
        str: a path to the directory
    """
    if domain:
        directory = os.path.join(directory, domain)

    # Expand and normalize the path
    directory = os.path.normpath(directory)
    directory = os.path.expanduser(directory)

    if not os.path.isdir(directory):
        try:
            save = os.umask(0)
            os.makedirs(directory, 0o777)  # give everyone rights to r/w here
        except OSError:
            LOG.warning("Failed to create: " + directory)
            pass
        finally:
            os.umask(save)

    return directory


def create_file(filename):
    """ Create the file filename and create any directories needed

        Args:
            filename: Path to the file to be created
    """
    try:
        os.makedirs(os.path.dirname(filename))
    except OSError:
        pass
    with open(filename, 'w') as f:
        f.write('')


def create_signal(signal_name, config=None):
    """Create a named signal

    Args:
        signal_name (str): The signal's name.  Must only contain characters
            valid in filenames.
        config (dict): mycroft.conf, to read ipc directory from
    """
    try:
        path = os.path.join(get_ipc_directory(config=config),
                            "signal", signal_name)
        create_file(path)
        return os.path.isfile(path)
    except IOError:
        return False


def check_for_signal(signal_name, sec_lifetime=0, config=None):
    """See if a named signal exists

    Args:
        signal_name (str): The signal's name.  Must only contain characters
            valid in filenames.
        sec_lifetime (int, optional): How many seconds the signal should
            remain valid.  If 0 or not specified, it is a single-use signal.
            If -1, it never expires.
        config (dict): mycroft.conf, to read ipc directory from

    Returns:
        bool: True if the signal is defined, False otherwise
    """
    path = os.path.join(get_ipc_directory(config=config),
                        "signal", signal_name)
    if os.path.isfile(path):
        if sec_lifetime == 0:
            # consume this single-use signal
            os.remove(path)
        elif sec_lifetime == -1:
            return True
        elif int(os.path.getctime(path) + sec_lifetime) < int(time.time()):
            # remove once expired
            os.remove(path)
            return False
        return True

    # No such signal exists
    return False



import subprocess
from distutils.spawn import find_executable
from ovos_utils.gui import is_gui_installed
from ovos_utils.log import LOG


class InputDeviceHelper:
    def __init__(self) -> None:
        self.libinput_devices_list = []
        self.xinput_devices_list = []
        if not find_executable("libinput") and not find_executable("xinput"):
            LOG.warning("Could not find libinput, input device detection will be inaccurate")

    # ToDo: add support for discovering the input device based of a connected
    # monitors, currently linux only supports input listing directly from the
    # system
    def _build_linput_devices_list(self):
        # Always clear the list before building it
        self.libinput_devices_list.clear()

        input_device_names = []
        input_device_kernel_path = []
        input_device_group = []
        input_device_capabilities = []

        # Get the list of devices from libinput
        proc_output = subprocess.check_output(['libinput', 'list-devices']).decode('utf-8')
        for line in proc_output.splitlines():
            if line.startswith('Device'):
                unformated_device_name = line.split(':')[1]
                unformated_device_name = unformated_device_name.strip()
                input_device_names.append(unformated_device_name)

        for line in proc_output.splitlines():
            if line.startswith('Kernel'):
                unformated_device_kernel_path = line.split(':')[1]
                unformated_device_kernel_path = unformated_device_kernel_path.strip()
                input_device_kernel_path.append(unformated_device_kernel_path)

        for line in proc_output.splitlines():
            if line.startswith('Group'):
                unformated_device_group = line.split(':')[1]
                unformated_device_group = unformated_device_group.strip()
                input_device_group.append(unformated_device_group)

        for line in proc_output.splitlines():
            if line.startswith("Capabilities"):
                unformated_device_capabilities = line.split(':')[1]
                unformated_device_capabilities = unformated_device_capabilities.strip()
                # check if there is a comma in the string and if so, split it
                if ' ' in unformated_device_capabilities:
                    unformated_device_capabilities = unformated_device_capabilities.split(
                        ' ')
                else:
                    unformated_device_capabilities = [
                        unformated_device_capabilities]

                input_device_capabilities.append(
                    unformated_device_capabilities)

        for i in range(len(input_device_names)):
            self.libinput_devices_list.append(
                {"Device": input_device_names[i],
                 "Kernel": input_device_kernel_path[i],
                 "Group": input_device_group[i],
                 "Capabilities": input_device_capabilities[i]
                 })

    def _get_libinput_devices_list(self):
        if find_executable("libinput"):
            try:
                self._build_linput_devices_list()
            except:
                self.libinput_devices_list.clear()
                LOG.exception("Failed to query libinput for devices")
        return self.libinput_devices_list

    def _build_xinput_devices_list(self):
        # Always clear the list before building it
        self.xinput_devices_list.clear()

        # Get the list of devices from xinput
        proc_output = subprocess.check_output(['xinput', 'list']).decode('utf-8')
        for line in proc_output.splitlines():
            # skip virtual devices always present
            if "↳" not in line or "XTEST" in line:
                continue
            line = line.replace("↳", "").replace("⎡", "").replace("⎣", "").replace("⎜", "").strip()

            if "pointer" in line:
                dev = {"Device": line.split("id=")[0].strip(),
                       "Capabilities": ["mouse"]}
                self.xinput_devices_list.append(dev)
            if "keyboard" in line:
                dev = {"Device": line.split("id=")[0].strip(),
                       "Capabilities": ["keyboard"]}
                self.xinput_devices_list.append(dev)

    def _get_xinput_devices_list(self):
        if find_executable("xinput"):
            try:
                self._build_xinput_devices_list()
            except:
                self.xinput_devices_list.clear()
                LOG.exception("Failed to query xinput for devices")
        return self.xinput_devices_list

    def get_input_device_list(self):
        # check if any of the devices support touch or mouse
        self._get_libinput_devices_list()
        self._get_xinput_devices_list()
        return self.libinput_devices_list + self.xinput_devices_list

    def can_use_touch_mouse(self):
        if not find_executable("libinput") and not find_executable("xinput"):
            # if gui installed assume we have a mouse
            # otherwise let's assume we are a server or something...
            return is_gui_installed()
        for device in self.get_input_device_list():
            if "touch" in device["Capabilities"] or \
                    "mouse" in device["Capabilities"] or \
                    "tablet" in device["Capabilities"] or \
                    "pointer" in device["Capabilities"] or \
                    "gesture" in device["Capabilities"]:
                return True
        return False

    def can_use_keyboard(self):
        for device in self.get_input_device_list():
            if "keyboard" in device["Capabilities"]:
                return True
        return False


def can_use_touch_mouse():
    return InputDeviceHelper().can_use_touch_mouse()


def can_use_keyboard():
    return InputDeviceHelper().can_use_keyboard()



import time

from ovos_utils.intents.intent_service_interface import IntentQueryApi
from ovos_utils.log import LOG
from ovos_utils.messagebus import FakeMessage as Message


class ConverseTracker:
    """ Using the messagebus this class recreates/keeps track of the state
    of the converse system, it uses both passive listening and active
    queries to sync it's state, it also emits 2 new bus events

    Implements https://github.com/MycroftAI/mycroft-core/pull/1468
    """
    bus = None
    active_skills = []
    converse_timeout = 5  # MAGIC NUMBER  hard coded in mycroft-core
    last_conversed = None
    intent_api = None

    @classmethod
    def connect_bus(cls, mycroft_bus):
        """Registers the bus object to use."""
        # PATCH - in mycroft-core this would be handled in intent_service
        # in here it is done in MycroftSkill.bind so i added this
        # conditional check
        if cls.bus is None and mycroft_bus is not None:
            cls.bus = mycroft_bus
            cls.intent_api = IntentQueryApi(cls.bus)
            cls.register_bus_events()

    @classmethod
    def register_bus_events(cls):
        cls.bus.on('active_skill_request', cls.handle_activate_request)
        cls.bus.on('skill.converse.response', cls.handle_converse_response)
        cls.bus.on("mycroft.skill.handler.start", cls.handle_intent_start)
        cls.bus.on("recognizer_loop:utterance", cls.handle_utterance)

    # public methods
    @classmethod
    def check_skill(cls, skill_id):
        """ Check if a skill is active """
        cls.filter_active_skills()
        for skill in list(cls.active_skills):
            if skill[0] == skill_id:
                return True
        return False

    @classmethod
    def filter_active_skills(cls):
        """ Removes expired skills from active skill list """
        # filter timestamps
        for skill in list(cls.active_skills):
            if time.time() - skill[1] <= cls.converse_timeout * 60:
                cls.remove_active_skill(skill[0])

    @classmethod
    def sync_with_intent_service(cls):
        """sync active skill list using intent api

        WARNING
            we don't have the timestamps so order might be messed up!!
            avoid calling this until
        """
        skill_ids = cls.intent_api.get_active_skills(include_timestamps=True)
        if skill_ids:
            if len(skill_ids[0]) == 2:
                # PR was merged! hurray!
                cls.active_skills = skill_ids
            else:
                # hoping they come sorted by timestamp....
                # older to newer (most recently used)
                for skill_id in reversed(skill_ids):
                    # are we tracking this skill ?
                    if not cls.check_skill(skill_id):
                        # we missed adding this skill in our tracking
                        cls.add_active_skill(skill_id)
                for skill in cls.active_skills:
                    if skill[0] not in skill_ids:
                        # we missed removing this skill in our tracking
                        cls.remove_active_skill(skill[0])

    # https://github.com/MycroftAI/mycroft-core/pull/1468
    @classmethod
    def remove_active_skill(cls, skill_id, silent=False):
        """
        Emits "converse.skill.deactivated" event, improvement of #1468
        """
        for skill in list(cls.active_skills):
            if skill[0] == skill_id:
                cls.active_skills.remove(skill)
                if not silent:
                    cls.bus.emit(Message("converse.skill.deactivated",
                                         {"skill_id": skill[0]}))

    @classmethod
    def add_active_skill(cls, skill_id):
        """
        Emits "converse.skill.activated" event, improvement of #1468
        """
        # search the list for an existing entry that already contains it
        # and remove that reference
        if skill_id != '':
            cls.remove_active_skill(skill_id, silent=True)
            # add skill with timestamp to start of skill_list
            cls.active_skills.insert(0, [skill_id, time.time()])
            # this might be sent more than once and it's perfectly fine
            # it's just a new info message not consumed anywhere by default
            cls.bus.emit(Message("converse.skill.activated",
                                 {"skill_id": skill_id}))
        else:
            LOG.warning('Skill ID was empty, won\'t add to list of '
                        'active skills.')

    # status tracking
    @classmethod
    def handle_activate_request(cls, message):
        """
        a skill bumped itself to the top of active skills list
        duplicate functionality from mycroft-core, keeping list in sync
        """
        skill_id = message.data["skill_id"]
        cls.add_active_skill(skill_id)

    @classmethod
    def handle_converse_error(cls, message):
        """
        a skill was removed from active skill list due to converse error
        duplicate functionality from mycroft-core, keeping list in sync
        """
        skill_id = message.data["skill_id"]
        if message.data["error"] == "skill id does not exist":
            cls.remove_active_skill(skill_id)

    @classmethod
    def handle_intent_start(cls, message):
        """
        duplicate functionality from mycroft-core, keeping list in sync

        TODO skill_id from message, core is not passing it along... it used
        to be possible to retrieve it from munged message but that changed.
        send a PR (if those got merged this code wouldn't exist)

        handle_utterance will take over this functionality for now
        handle_converse_response will take corrective action
        """
        # skill_id = message.data["skill_id"]
        # bump skill to top of active list
        # cls.add_active_skill(skill_id)

    @classmethod
    def handle_utterance(cls, message):
        """
        duplicate functionality from mycroft-core, keeping list in sync

        WORKAROUND - skill_id missing in handle_intent_start, will keep list
        in sync by using the IntentAPI to check what skill the utterance
        should trigger

        handle_converse_response will take corrective action
        """
        # NOTE borked in mycroft-core
        # needs https://github.com/MycroftAI/mycroft-core/pull/2786
        skill_id = cls.intent_api.get_skill(message.data["utterances"][0])
        if skill_id:
            # this skill will trigger and therefore is the last active skill
            cls.add_active_skill(skill_id)
        # will remove expired intents from list
        cls.filter_active_skills()

    @classmethod
    def handle_converse_response(cls, message):
        """
        tracks last_conversed skill

        FAILSAFE - additional checks to correct active skills list,
        but that should never happen, accounts for mistakes in
        handle_utterance / intent_api

        accounts for https://github.com/MycroftAI/mycroft-core/pull/2786
        not yet being merged
        """
        skill_id = message.data["skill_id"]
        if 'error' in message.data:
            cls.handle_converse_error(message)
        elif message.data.get('result') is True:
            cls.last_conversed = skill_id
            if not cls.check_skill(skill_id):
                # seems like we missed adding this skill to active list
                # NOTE this is a failsafe and should never trigger
                # since this answered True we have the real timestamp
                cls.add_active_skill(skill_id)
        elif not cls.check_skill(skill_id):
            # seems like we missed adding this skill to active list
            # NOTE this is a failsafe and should never trigger
            # since this answered false and we don't have the real timestamp
            # let's add it to the end of the active_skills list
            ts = time.time()
            if len(cls.active_skills):
                ts = cls.active_skills[-1][1]
            cls.active_skills.append([skill_id, ts])



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import SMTP_SSL

from ovos_utils.log import LOG


def send_smtp(user, pswd, sender,
              destinatary, subject, contents,
              host, port=465):
    with SMTP_SSL(host=host, port=port) as server:
        server.login(user, pswd)
        msg = MIMEMultipart()
        msg['From'] = sender
        msg['To'] = destinatary
        msg['Subject'] = subject
        msg.attach(MIMEText(contents))
        server.sendmail(sender, destinatary, msg.as_string())


def send_email(subject, body, recipient=None):
    try:
        from ovos_config.config import read_mycroft_config
        config = read_mycroft_config()
    except ImportError:
        LOG.warning("Config not provided and ovos_config not available")
        config = dict()
    mail_config = config.get("email") or {}
    if not mail_config:
        raise KeyError("email configuration not set")

    smtp_config = mail_config["smtp"]
    user = smtp_config["username"]
    pswd = smtp_config["password"]
    host = smtp_config["host"]
    port = smtp_config.get("port", 465)

    recipient = recipient or mail_config.get("recipient") or user

    send_smtp(user, pswd,
              user, recipient,
              subject, body,
              host, port)



try:
    import alsaaudio
except ImportError:
    alsaaudio = None
from ovos_utils.log import LOG


class AlsaControl:
    _mixer = None

    def __init__(self, control=None):
        # TODO: Deprecate class in 0.1
        LOG.warning(f"This class is deprecated! Controls moved to"
                    f"ovos_phal_plugin_alsa.AlsaVolumeControlPlugin")
        if alsaaudio is None:
            LOG.error("pyalsaaudio not installed")
            LOG.info("Run pip install pyalsaaudio==0.8.2")
            raise ImportError
        if control is None:
            control = alsaaudio.mixers()[0]
        self.get_mixer(control)

    @property
    def mixer(self):
        return self._mixer

    def get_mixer(self, control="Master"):
        if self._mixer is None:
            try:
                mixer = alsaaudio.Mixer(control)
            except Exception as e:
                try:
                    mixer = alsaaudio.Mixer(control)
                except Exception as e:
                    try:
                        if control != "Master":
                            LOG.warning("could not allocate requested mixer, "
                                        "falling back to 'Master'")
                            mixer = alsaaudio.Mixer("Master")
                        else:
                            raise
                    except Exception as e:
                        LOG.error("Couldn't allocate mixer")
                        LOG.exception(e)
                        raise
            self._mixer = mixer
        return self.mixer

    def increase_volume(self, percent):
        volume = self.get_volume()
        if isinstance(volume, list):
            volume = volume[0]
        volume += percent
        if volume < 0:
            volume = 0
        elif volume > 100:
            volume = 100
        self.mixer.setvolume(int(volume))

    def decrease_volume(self, percent):
        volume = self.get_volume()
        if isinstance(volume, list):
            volume = volume[0]
        volume -= percent
        if volume < 0:
            volume = 0
        elif volume > 100:
            volume = 100
        self.mixer.setvolume(int(volume))

    def set_volume_percent(self, percent):
        self.set_volume(percent)

    def set_volume(self, volume):
        if volume < 0:
            volume = 0
        elif volume > 100:
            volume = 100
        self.mixer.setvolume(int(volume))

    def volume_range(self):
        return self.mixer.getrange()

    def is_muted(self):
        return bool(self.mixer.getmute()[0])

    def mute(self):
        return self.mixer.setmute(1)

    def unmute(self):
        return self.mixer.setmute(0)

    def toggle_mute(self):
        if self.is_muted():
            self.unmute()
        else:
            self.mute()

    def get_volume(self):
        return self.mixer.getvolume()[0]

    def get_volume_percent(self):
        return self.get_volume()



import subprocess
import re
import collections
from ovos_utils.log import LOG


class PulseAudio:
    volume_re = re.compile('^set-sink-volume ([^ ]+) (.*)')
    mute_re = re.compile('^set-sink-mute ([^ ]+) ((?:yes)|(?:no))')

    def __init__(self):
        # TODO: Deprecate class in 0.1
        LOG.warning(f"This class is deprecated! Controls moved to"
                    f"ovos_phal_plugin_pulseaudio.PulseAudioVolumeControlPlugin")
        self._mute = collections.OrderedDict()
        self._volume = collections.OrderedDict()
        self.update()

    def normalize_sinks(self):
        self.unmute_all()
        volume = self.get_volume()
        self.set_all_volumes(volume)

    def update(self):
        proc = subprocess.Popen(['pacmd', 'dump'], stdout=subprocess.PIPE)

        for line in proc.stdout:
            line = line.decode("utf-8")
            volume_match = PulseAudio.volume_re.match(line)
            mute_match = PulseAudio.mute_re.match(line)

            if volume_match:
                self._volume[volume_match.group(1)] = int(
                    volume_match.group(2), 16)
            elif mute_match:
                self._mute[mute_match.group(1)] = mute_match.group(
                    2).lower() == "yes"

    def _vol_to_percent(self, vol):
        max_vol = 65536
        percent = vol * 100 / max_vol
        return percent

    def _percent_to_vol(self, percent):
        max_vol = 65536
        vol = percent * max_vol / 100
        return vol

    def get_volume_percent(self, sink=None):
        vol = self.get_sink_volume(sink)
        return self._vol_to_percent(vol)

    def get_mute(self, sink=None):
        if not sink:
            sink = list(self._mute.keys())[0]

        return self._mute[sink]

    def get_volume(self, sink=None):
        return self.get_sink_volume(sink)

    def get_sink_volume(self, sink=None):
        if not sink:
            sink = list(self._volume.keys())[0]

        return self._volume[sink]

    def set_mute(self, mute, sink=None):
        if not sink:
            sink = list(self._mute.keys())[0]

        subprocess.Popen(
            ['pacmd', 'set-sink-mute', sink, 'yes' if mute else 'no'],
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self._mute[sink] = mute

    def set_volume(self, volume, sink=None):
        self.set_sink_volume(volume, sink)

    def set_volume_percent(self, volume, sink=None):
        self.set_sink_volume(self._percent_to_vol(volume), sink)

    def set_sink_volume(self, volume, sink=None):
        if not sink:
            sink = list(self._volume.keys())[0]
        volume = int(volume)
        subprocess.Popen(['pacmd', 'set-sink-volume', sink, hex(volume)],
                         stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self._volume[sink] = volume

    def mute_all(self):
        for sink in self.list_sinks():
            self.set_mute(True, sink)

    def unmute_all(self):
        for sink in self.list_sinks():
            self.set_mute(False, sink)

    def set_all_volumes(self, volume):
        self.set_all_sink_volumes(volume)

    def get_all_volumes(self):
        return self.get_all_sink_volumes()

    def set_all_volumes_percent(self, percent):
        volume = self._percent_to_vol(percent)
        self.set_all_sink_volumes(volume)

    def get_all_volumes_percent(self):
        return [self._vol_to_percent(volume) for volume in
                self.get_all_sink_volumes()]

    def set_all_sink_volumes(self, volume):
        for sink in self.list_sinks():
            self.set_volume(volume, sink)

    def get_all_sink_volumes(self):
        volumes = []
        for sink in self.list_sinks():
            volumes.append(self.get_volume(sink))
        return volumes

    def list_sinks(self):
        proc = subprocess.Popen(['pacmd', 'list-sinks'],
                                stdout=subprocess.PIPE)
        sinks = []
        for line in proc.stdout:
            line = line.decode("utf-8").strip()
            if line.startswith("name: <"):
                sink = line.replace("name: <", "")[:-1]
                sinks.append(sink)
        return sinks

    def list_sources(self):
        proc = subprocess.Popen(['pacmd', 'list-sources'],
                                stdout=subprocess.PIPE)
        sinks = []
        for line in proc.stdout:
            line = line.decode("utf-8").strip()
            if line.startswith("name: <"):
                sink = line.replace("name: <", "")[:-1]
                sinks.append(sink)
        return sinks

    def increase_volume(self, percent):
        volume = self.get_volume_percent()
        volume += percent
        if volume < 0:
            volume = 0
        elif volume > 100:
            volume = 100
        self.set_all_volumes_percent(volume)

    def decrease_volume(self, percent):
        volume = self.get_volume_percent()
        volume -= percent
        if volume < 0:
            volume = 0
        elif volume > 100:
            volume = 100
        self.set_all_volumes_percent(volume)



import json
from os import makedirs
from os.path import join, expanduser, exists, isfile

import ovos_utils.xdg_utils as xdg
from ovos_utils.log import LOG


# TODO - deprecate this submodule in 0.1.0
# note that a couple of these are also used inside ovos-utils
# perhaps those usages should also move into workshop ?

def get_default_lang():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locale import get_default_lang as _get
        return _get()
    except ImportError:
        return read_mycroft_config().get("lang", "en-us")


def find_user_config():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locations import find_user_config as _get
        return _get()
    except ImportError:

        return join(get_xdg_config_save_path(), get_config_filename())


def get_webcache_location():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    return join(get_xdg_config_save_path(), 'web_cache.json')


def get_config_locations(default=True, web_cache=True, system=True,
                         old_user=True, user=True):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locations import get_config_locations as _get
        return _get(default, web_cache, system, old_user, user)
    except ImportError:
        locs = []
        ovos_cfg = get_ovos_config()
        if default:
            locs.append(ovos_cfg["default_config_path"])
        if system:
            locs.append(f"/etc/{ovos_cfg['base_folder']}/{ovos_cfg['config_filename']}")
        if web_cache:
            locs.append(get_webcache_location())
        if old_user:
            locs.append(f"~/.{ovos_cfg['base_folder']}/{ovos_cfg['config_filename']}")
        if user:
            locs.append(f"{get_xdg_config_save_path()}/{ovos_cfg['config_filename']}")
        return locs


def get_ovos_config():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.meta import get_ovos_config as _get
        return _get()
    except ImportError:
        return {"xdg": True,
                "base_folder": "mycroft",
                "config_filename": "mycroft.conf"}


def get_xdg_base():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.meta import get_xdg_base as _get
        return _get()
    except ImportError:
        return "mycroft"


def get_xdg_config_locations():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    # This includes both the user config and
    # /etc/xdg/mycroft/mycroft.conf
    xdg_paths = list(reversed(
        [join(p, get_config_filename())
         for p in get_xdg_config_dirs()]
    ))
    return xdg_paths


def get_xdg_data_dirs():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locations import get_xdg_data_dirs as _get
        return _get()
    except ImportError:
        return [expanduser("~/.local/share/mycroft")]


def get_xdg_config_dirs(folder=None):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locations import get_xdg_config_dirs as _get
        return _get()
    except ImportError:
        folder = folder or get_xdg_base()
        xdg_dirs = xdg.xdg_config_dirs() + [xdg.xdg_config_home()]
        return [join(path, folder) for path in xdg_dirs]


def get_xdg_cache_save_path(folder=None):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locations import get_xdg_cache_save_path as _get
        return _get()
    except ImportError:
        folder = folder or get_xdg_base()
        return join(xdg.xdg_cache_home(), folder)


def get_xdg_data_save_path():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locations import get_xdg_data_save_path as _get
        return _get()
    except ImportError:
        return expanduser("~/.local/share/mycroft")


def get_xdg_config_save_path():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locations import get_xdg_config_save_path as _get
        return _get()
    except ImportError:
        return expanduser("~/.config/mycroft")


def is_using_xdg():
    """ DEPRECATED """
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    return True


def set_xdg_base(*args, **kwargs):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.meta import set_xdg_base as _set
        _set(*args, **kwargs)
    except:
        pass


def set_config_filename(*args, **kwargs):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.meta import config_filename as _set
        _set(*args, **kwargs)
    except:
        pass


def get_config_filename():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.locale import get_config_filename as _get
        return _get()
    except ImportError:
        return "mycroft.conf"


def get_ovos_default_config_paths():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.meta import get_ovos_default_config_paths as _get
        return _get()
    except:
        return ["/etc/OpenVoiceOS/ovos.conf"]


def read_mycroft_config():
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config import Configuration
        return Configuration()
    except ImportError:
        pass
    path = expanduser(f"~/.config/mycroft/mycroft.conf")
    if isfile(path):
        with open(path) as f:
            return json.load(f)
    return {
        # TODO - default cfg
        "lang": "en-us"
    }


def update_mycroft_config(config, path=None, bus=None):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.config import update_mycroft_config as _update
        _update(config, path, bus)
    except ImportError:
        pass
    # save in default user location
    path = expanduser(f"~/.config/mycroft")
    makedirs(path, exist_ok=True)
    with open(f"{path}/mycroft.conf", "w") as f:
        json.dump(config, f, indent=2)


def set_default_config(*args, **kwargs):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.meta import set_default_config as _set
        _set(*args, **kwargs)
    except:
        pass


def save_ovos_core_config(*args, **kwargs):
    LOG.warning("configuration moved to the `ovos_config` package. This submodule "
                "will be removed in ovos_utils 0.1.0")
    try:
        from ovos_config.meta import save_ovos_config as _set
        _set(*args, **kwargs)
    except:
        pass


try:
    from ovos_config.models import (
        LocalConf,
        ReadOnlyConfig,
        MycroftUserConfig,
        MycroftDefaultConfig,
        MycroftSystemConfig,
        MycroftXDGConfig
    )
except ImportError:
    LOG.warning("configuration classes moved to the `ovos_config.models` package. "
                "This submodule will be removed in ovos_utils 0.1.0")
    from combo_lock import NamedLock
    import yaml
    from ovos_utils.json_helper import load_commented_json, merge_dict


    class LocalConf(dict):
        """Config dictionary from file."""
        allow_overwrite = True
        # lock is shared among all subclasses,
        # regardless of what file is being edited only one file should change at a time
        # this ensure orderly behaviour in anything monitoring changes,
        #   eg FileWatcher util, configuration.patch bus handlers
        __lock = NamedLock("ovos_config")

        def __init__(self, path):
            super().__init__(self)
            self.path = path
            if path:
                self.load_local(path)

        def _get_file_format(self, path=None):
            """The config file format
            supported file extensions:
            - json (.json)
            - commented json (.conf)
            - yaml (.yaml/.yml)

            returns "yaml" or "json"
            """
            path = path or self.path
            if not path:
                return "dict"
            if path.endswith(".yml") or path.endswith(".yaml"):
                return "yaml"
            else:
                return "json"

        def load_local(self, path=None):
            """Load local json file into self.

            Args:
                path (str): file to load
            """
            path = path or self.path
            if not path:
                LOG.error(f"in memory configuration, nothing to load")
                return
            if exists(path) and isfile(path):
                with self.__lock:
                    try:
                        if self._get_file_format(path) == "yaml":
                            with open(path) as f:
                                config = yaml.safe_load(f)
                        else:
                            config = load_commented_json(path)
                        if config:
                            for key in config:
                                self.__setitem__(key, config[key])
                            LOG.debug(f"Configuration {path} loaded")
                        else:
                            LOG.debug(f"Empty config found at: {path}")
                    except Exception as e:
                        LOG.exception(f"Error loading configuration '{path}'")
            else:
                LOG.debug(f"Configuration '{path}' not defined, skipping")

        def reload(self):
            self.load_local(self.path)

        def store(self, path=None):
            path = path or self.path
            if not path:
                LOG.error(f"in memory configuration, no save location")
                return
            with self.__lock:
                if self._get_file_format(path) == "yaml":
                    with open(path, 'w+') as f:
                        yaml.dump(dict(self), f, allow_unicode=True,
                                  default_flow_style=False, sort_keys=False)
                else:
                    with open(path, 'w+') as f:
                        json.dump(self, f, indent=2)

        def merge(self, conf):
            merge_dict(self, conf)


    class ReadOnlyConfig(LocalConf):
        """ read only  """

        def __init__(self, path, allow_overwrite=False):
            super().__init__(path)
            self.allow_overwrite = allow_overwrite

        def reload(self):
            old = self.allow_overwrite
            self.allow_overwrite = True
            super().reload()
            self.allow_overwrite = old

        def __setitem__(self, key, value):
            if not self.allow_overwrite:
                raise PermissionError(f"{self.path} is read only! it can not be modified at runtime")
            super().__setitem__(key, value)

        def merge(self, *args, **kwargs):
            if not self.allow_overwrite:
                raise PermissionError(f"{self.path} is read only! it can not be modified at runtime")
            super().merge(*args, **kwargs)

        def store(self, path=None):
            if not self.allow_overwrite:
                raise PermissionError(f"{self.path} is read only! it can not be modified at runtime")
            super().store(path)


    class MycroftDefaultConfig(ReadOnlyConfig):
        def __init__(self):
            super().__init__(join(get_xdg_config_save_path(), get_config_filename()))

        def set_root_config_path(self, root_config):
            # in case we got it wrong / non standard
            self.path = root_config
            self.reload()


    class MycroftSystemConfig(ReadOnlyConfig):
        def __init__(self, allow_overwrite=False):
            super().__init__("/etc/mycroft/mycroft.conf", allow_overwrite)


    class RemoteConf(LocalConf):
        def __init__(self, cache=get_webcache_location()):
            super(RemoteConf, self).__init__(cache)


    MycroftXDGConfig = MycroftUserConfig = MycroftDefaultConfig



from ovos_utils.intents.intent_service_interface import IntentQueryApi, \
    IntentServiceInterface
from ovos_utils.intents.converse import ConverseTracker
from ovos_utils.intents.layers import IntentLayers

try:
    from adapt.intent import IntentBuilder, Intent
except ImportError:
    # adapt is optional, these classes are mainly syntactic sugar

    class Intent:
        def __init__(self, name, requires, at_least_one, optional):
            """Create Intent object
            Args:
                name(str): Name for Intent
                requires(list): Entities that are required
                at_least_one(list): One of these Entities are required
                optional(list): Optional Entities used by the intent
            """
            self.name = name
            self.requires = requires
            self.at_least_one = at_least_one
            self.optional = optional

        def validate(self, tags, confidence):
            """Using this method removes tags from the result of validate_with_tags
            Returns:
                intent(intent): Results from validate_with_tags
            """
            raise NotImplementedError("please install adapt-parser")

        def validate_with_tags(self, tags, confidence):
            """Validate whether tags has required entites for this intent to fire
            Args:
                tags(list): Tags and Entities used for validation
                confidence(float): The weight associate to the parse result,
                    as indicated by the parser. This is influenced by a parser
                    that uses edit distance or context.
            Returns:
                intent, tags: Returns intent and tags used by the intent on
                    failure to meat required entities then returns intent with
                    confidence
                    of 0.0 and an empty list for tags.
            """
            raise NotImplementedError("please install adapt-parser")


    class IntentBuilder:
        """
        IntentBuilder, used to construct intent parsers.
        Attributes:
            at_least_one(list): A list of Entities where one is required.
                These are separated into lists so you can have one of (A or B) and
                then require one of (D or F).
            requires(list): A list of Required Entities
            optional(list): A list of optional Entities
            name(str): Name of intent
        Notes:
            This is designed to allow construction of intents in one line.
        Example:
            IntentBuilder("Intent")\
                .requires("A")\
                .one_of("C","D")\
                .optional("G").build()
        """

        def __init__(self, intent_name):
            """
            Constructor
            Args:
                intent_name(str): the name of the intents that this parser
                parses/validates
            """
            self.at_least_one = []
            self.requires = []
            self.optional = []
            self.name = intent_name

        def one_of(self, *args):
            """
            The intent parser should require one of the provided entity types to
            validate this clause.
            Args:
                args(args): *args notation list of entity names
            Returns:
                self: to continue modifications.
            """
            self.at_least_one.append(args)
            return self

        def require(self, entity_type, attribute_name=None):
            """
            The intent parser should require an entity of the provided type.
            Args:
                entity_type(str): an entity type
                attribute_name(str): the name of the attribute on the parsed intent.
                Defaults to match entity_type.
            Returns:
                self: to continue modifications.
            """
            if not attribute_name:
                attribute_name = entity_type
            self.requires += [(entity_type, attribute_name)]
            return self

        def optionally(self, entity_type, attribute_name=None):
            """
            Parsed intents from this parser can optionally include an entity of the
             provided type.
            Args:
                entity_type(str): an entity type
                attribute_name(str): the name of the attribute on the parsed intent.
                Defaults to match entity_type.
            Returns:
                self: to continue modifications.
            """
            if not attribute_name:
                attribute_name = entity_type
            self.optional += [(entity_type, attribute_name)]
            return self

        def build(self):
            """
            Constructs an intent from the builder's specifications.
            :return: an Intent instance.
            """
            return Intent(self.name, self.requires,
                          self.at_least_one, self.optional)


class AdaptIntent(IntentBuilder):
    """Wrapper for IntentBuilder setting a blank name.

    Args:
        name (str): Optional name of intent
    """

    def __init__(self, name=''):
        super().__init__(name)




import json
import time
from copy import deepcopy
from inspect import signature
from threading import Event

# from ovos_utils.configuration import read_mycroft_config, get_default_lang
from pyee import BaseEventEmitter

from ovos_utils import create_loop
from ovos_utils.json_helper import merge_dict
from ovos_utils.log import LOG
from ovos_utils.metrics import Stopwatch

_DEFAULT_WS_CONFIG = {"host": "0.0.0.0",
                      "port": 8181,
                      "route": "/core",
                      "ssl": False}


def dig_for_message():
    try:
        from ovos_bus_client.message import dig_for_message as _dig
        return _dig()
    except ImportError:
        pass
    return None


class FakeBus:
    def __init__(self, *args, **kwargs):
        self.started_running = False
        self.ee = kwargs.get("emitter") or BaseEventEmitter()
        self.ee.on("error", self.on_error)
        self.on_open()

    def on(self, msg_type, handler):
        self.ee.on(msg_type, handler)

    def once(self, msg_type, handler):
        self.ee.once(msg_type, handler)

    def emit(self, message):
        self.ee.emit("message", message.serialize())
        self.ee.emit(message.msg_type, message)

    def wait_for_message(self, message_type, timeout=3.0):
        """Wait for a message of a specific type.

        Arguments:
            message_type (str): the message type of the expected message
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """
        received_event = Event()
        received_event.clear()

        msg = None

        def rcv(m):
            nonlocal msg
            msg = m
            received_event.set()

        self.ee.once(message_type, rcv)
        received_event.wait(timeout)
        return msg

    def wait_for_response(self, message, reply_type=None, timeout=3.0):
        """Send a message and wait for a response.

        Arguments:
            message (Message): message to send
            reply_type (str): the message type of the expected reply.
                              Defaults to "<message.msg_type>.response".
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """
        reply_type = reply_type or message.msg_type + ".response"
        received_event = Event()
        received_event.clear()

        msg = None

        def rcv(m):
            nonlocal msg
            msg = m
            received_event.set()

        self.ee.once(reply_type, rcv)
        self.emit(message)
        received_event.wait(timeout)
        return msg

    def remove(self, msg_type, handler):
        try:
            self.ee.remove_listener(msg_type, handler)
        except:
            pass

    def remove_all_listeners(self, event_name):
        self.ee.remove_all_listeners(event_name)

    def create_client(self):
        return self

    def on_error(self, error):
        LOG.error(error)

    def on_open(self):
        pass

    def on_close(self):
        pass

    def run_forever(self):
        self.started_running = True

    def run_in_thread(self):
        self.run_forever()

    def close(self):
        self.on_close()


class _MutableMessage(type):
    """ To override isinstance checks we need to use a metaclass """

    def __instancecheck__(self, instance):
        try:
            from ovos_bus_client.message import Message as _MycroftMessage
            if isinstance(instance, _MycroftMessage):
                return True
        except:
            pass
        try:
            from mycroft_bus_client.message import Message as _MycroftMessage
            if isinstance(instance, _MycroftMessage):
                return True
        except:
            pass
        return super().__instancecheck__(instance)


# fake Message object to allow usage without ovos-bus-client installed
class FakeMessage(metaclass=_MutableMessage):
    """ fake Message object to allow usage with FakeBus without ovos-bus-client installed"""

    def __new__(cls, *args, **kwargs):
        try:  # most common case
            from ovos_bus_client import Message as _M
            return _M(*args, **kwargs)
        except:
            pass
        try:  # some old install that upgraded during migration period
            from mycroft_bus_client import Message as _M
            return _M(*args, **kwargs)
        except:  # FakeMessage
            return super().__new__(cls, *args, **kwargs)

    def __init__(self, msg_type, data=None, context=None):
        """Used to construct a message object

        Message objects will be used to send information back and forth
        between processes of mycroft service, voice, skill and cli
        """
        self.msg_type = msg_type
        self.data = data or {}
        self.context = context or {}

    def __eq__(self, other):
        try:
            return other.msg_type == self.msg_type and \
                other.data == self.data and \
                other.context == self.context
        except:
            return False

    def serialize(self):
        """This returns a string of the message info.

        This makes it easy to send over a websocket. This uses
        json dumps to generate the string with type, data and context

        Returns:
            str: a json string representation of the message.
        """
        return json.dumps({'type': self.msg_type,
                           'data': self.data,
                           'context': self.context})

    @staticmethod
    def deserialize(value):
        """This takes a string and constructs a message object.

        This makes it easy to take strings from the websocket and create
        a message object.  This uses json loads to get the info and generate
        the message object.

        Args:
            value(str): This is the json string received from the websocket

        Returns:
            FakeMessage: message object constructed from the json string passed
            int the function.
            value(str): This is the string received from the websocket
        """
        obj = json.loads(value)
        return FakeMessage(obj.get('type') or '',
                           obj.get('data') or {},
                           obj.get('context') or {})

    def forward(self, msg_type, data=None):
        """ Keep context and forward message

        This will take the same parameters as a message object but use
        the current message object as a reference.  It will copy the context
        from the existing message object.

        Args:
            msg_type (str): type of message
            data (dict): data for message

        Returns:
            FakeMessage: Message object to be used on the reply to the message
        """
        data = data or {}
        return FakeMessage(msg_type, data, context=self.context)

    def reply(self, msg_type, data=None, context=None):
        """Construct a reply message for a given message

        This will take the same parameters as a message object but use
        the current message object as a reference.  It will copy the context
        from the existing message object and add any context passed in to
        the function.  Check for a destination passed in to the function from
        the data object and add that to the context as a destination.  If the
        context has a source then that will be swapped with the destination
        in the context.  The new message will then have data passed in plus the
        new context generated.

        Args:
            msg_type (str): type of message
            data (dict): data for message
            context: intended context for new message

        Returns:
            FakeMessage: Message object to be used on the reply to the message
        """
        data = deepcopy(data) or {}
        context = context or {}

        new_context = deepcopy(self.context)
        for key in context:
            new_context[key] = context[key]
        if 'destination' in data:
            new_context['destination'] = data['destination']
        if 'source' in new_context and 'destination' in new_context:
            s = new_context['destination']
            new_context['destination'] = new_context['source']
            new_context['source'] = s
        return FakeMessage(msg_type, data, context=new_context)

    def response(self, data=None, context=None):
        """Construct a response message for the message

        Constructs a reply with the data and appends the expected
        ".response" to the message

        Args:
            data (dict): message data
            context (dict): message context
        Returns
            (Message) message with the type modified to match default response
        """
        return self.reply(self.msg_type + '.response', data, context)

    def publish(self, msg_type, data, context=None):
        """
        Copy the original context and add passed in context.  Delete
        any target in the new context. Return a new message object with
        passed in data and new context.  Type remains unchanged.

        Args:
            msg_type (str): type of message
            data (dict): date to send with message
            context: context added to existing context

        Returns:
            FakeMessage: Message object to publish
        """
        context = context or {}
        new_context = self.context.copy()
        for key in context:
            new_context[key] = context[key]

        if 'target' in new_context:
            del new_context['target']

        return FakeMessage(msg_type, data, context=new_context)


class Message(FakeMessage):
    def __int__(self, *args, **kwargs):
        LOG.warning(f"This reference is deprecated, import from "
                    f"`ovos_bus_client.message` directly")
        FakeMessage.__init__(self, *args, **kwargs)


def get_message_lang(message=None):
    """Get the language from the message or the default language.
    Args:
        message: message to check for language code.
    Returns:
        The language code from the message or the default language.
    """
    try:
        from ovos_config.locale import get_default_lang
        default_lang = get_default_lang()
    except ImportError:
        LOG.warning("ovos_config not available. Using default lang en-us")
        default_lang = "en-us"
    message = message or dig_for_message()
    if not message:
        return default_lang
    lang = message.data.get("lang") or message.context.get("lang") or default_lang
    return lang.lower()


def get_websocket(host, port, route='/', ssl=False, threaded=True):
    """
    Returns a connection to a websocket
    """
    from ovos_bus_client import MessageBusClient

    client = MessageBusClient(host, port, route, ssl)
    if threaded:
        client.run_in_thread()
    return client


def get_mycroft_bus(host: str = None, port: int = None, route: str = None,
                    ssl: bool = None):
    """
    Returns a connection to the mycroft messagebus
    """
    try:
        from ovos_config.config import read_mycroft_config
        config = read_mycroft_config().get('websocket') or dict()
    except ImportError:
        LOG.warning("ovos_config not available. Falling back to default WS")
        config = dict()
    host = host or config.get('host') or _DEFAULT_WS_CONFIG['host']
    port = port or config.get('port') or _DEFAULT_WS_CONFIG['port']
    route = route or config.get('route') or _DEFAULT_WS_CONFIG['route']
    if ssl is None:
        ssl = config.get('ssl') if 'ssl' in config else \
            _DEFAULT_WS_CONFIG['ssl']
    return get_websocket(host, port, route, ssl)


def listen_for_message(msg_type, handler, bus=None):
    """
    Continuously listens and reacts to a specific messagetype on the mycroft messagebus

    NOTE: when finished you should call bus.remove(msg_type, handler)
    """
    bus = bus or get_mycroft_bus()
    bus.on(msg_type, handler)
    return bus


def listen_once_for_message(msg_type, handler, bus=None):
    """
    listens and reacts once to a specific messagetype on the mycroft messagebus
    """
    auto_close = bus is None
    bus = bus or get_mycroft_bus()

    def _handler(message):
        handler(message)
        if auto_close:
            bus.close()

    bus.once(msg_type, _handler)
    return bus


def wait_for_reply(message, reply_type=None, timeout=3.0, bus=None):
    """Send a message and wait for a response.

    Args:
        message (FakeMessage or str or dict): message object or type to send
        reply_type (str): the message type of the expected reply.
                          Defaults to "<message.type>.response".
        timeout: seconds to wait before timeout, defaults to 3
    Returns:
        The received message or None if the response timed out
    """
    auto_close = bus is None
    bus = bus or get_mycroft_bus()
    if isinstance(message, str):
        try:
            message = json.loads(message)
        except:
            pass
    if isinstance(message, str):
        message = FakeMessage(message)
    elif isinstance(message, dict):
        message = FakeMessage(message["type"],
                              message.get("data"),
                              message.get("context"))
    elif not isinstance(message, FakeMessage):
        raise ValueError
    response = bus.wait_for_response(message, reply_type, timeout)
    if auto_close:
        bus.close()
    return response


def send_message(message, data=None, context=None, bus=None):
    auto_close = bus is None
    bus = bus or get_mycroft_bus()
    if isinstance(message, str):
        if isinstance(data, dict) or isinstance(context, dict):
            message = FakeMessage(message, data, context)
        else:
            try:
                message = json.loads(message)
            except:
                message = FakeMessage(message)
    if isinstance(message, dict):
        message = FakeMessage(message["type"],
                              message.get("data"),
                              message.get("context"))
    if not isinstance(message, FakeMessage):
        raise ValueError
    bus.emit(message)
    if auto_close:
        bus.close()


def send_binary_data_message(binary_data, msg_type="mycroft.binary.data",
                             msg_data=None, msg_context=None, bus=None):
    msg_data = msg_data or {}
    msg = {
        "type": msg_type,
        "data": merge_dict(msg_data, {"binary": binary_data.hex()}),
        "context": msg_context or None
    }
    send_message(msg, bus=bus)


def send_binary_file_message(filepath, msg_type="mycroft.binary.file",
                             msg_context=None, bus=None):
    with open(filepath, 'rb') as f:
        binary_data = f.read()
    msg_data = {"path": filepath}
    send_binary_data_message(binary_data, msg_type=msg_type, msg_data=msg_data,
                             msg_context=msg_context, bus=bus)


def decode_binary_message(message):
    if isinstance(message, str):
        try:  # json string
            message = json.loads(message)
            binary_data = message.get("binary") or message["data"]["binary"]
        except:  # hex string
            binary_data = message
    elif isinstance(message, dict):
        # data field or serialized message
        binary_data = message.get("binary") or message["data"]["binary"]
    else:
        # message object
        binary_data = message.data["binary"]
    # decode hex string
    return bytearray.fromhex(binary_data)


def to_alnum(skill_id):
    """Convert a skill id to only alphanumeric characters

     Non alpha-numeric characters are converted to "_"

    Args:
        skill_id (str): identifier to be converted
    Returns:
        (str) String of letters
    """
    return ''.join(c if c.isalnum() else '_' for c in str(skill_id))


def unmunge_message(message, skill_id):
    """Restore message keywords by removing the Letterified skill ID.
    Args:
        message (FakeMessage): Intent result message
        skill_id (str): skill identifier
    Returns:
        Message without clear keywords
    """
    if hasattr(message, "data") and isinstance(message.data, dict):
        skill_id = to_alnum(skill_id)
        for key in list(message.data.keys()):
            if key.startswith(skill_id):
                # replace the munged key with the real one
                new_key = key[len(skill_id):]
                message.data[new_key] = message.data.pop(key)

    return message


def get_handler_name(handler):
    """Name (including class if available) of handler function.

    Args:
        handler (function): Function to be named

    Returns:
        string: handler name as string
    """
    if '__self__' in dir(handler) and 'name' in dir(handler.__self__):
        return handler.__self__.name + '.' + handler.__name__
    else:
        return handler.__name__


def create_wrapper(handler, skill_id, on_start, on_end, on_error):
    """Create the default skill handler wrapper.

    This wrapper handles things like metrics, reporting handler start/stop
    and errors.
        handler (callable): method/function to call
        skill_id: skill_id for associated skill
        on_start (function): function to call before executing the handler
        on_end (function): function to call after executing the handler
        on_error (function): function to call for error reporting
    """

    def wrapper(message):
        stopwatch = Stopwatch()
        try:
            message = unmunge_message(message, skill_id)
            if on_start:
                on_start(message)

            with stopwatch:
                if len(signature(handler).parameters) == 0:
                    handler()
                else:
                    handler(message)

        except Exception as e:
            if on_error:
                if len(signature(on_error).parameters) == 2:
                    on_error(e, message)
                else:
                    on_error(e)
        finally:
            if on_end:
                on_end(message)

            # Send timing metrics
            context = message.context
            if context and 'ident' in context:
                try:
                    from mycroft.metrics import report_timing
                    report_timing(context['ident'], 'skill_handler', stopwatch,
                                  {'handler': handler.__name__,
                                   'skill_id': skill_id})
                except ImportError:
                    pass

    return wrapper


def create_basic_wrapper(handler, on_error=None):
    """Create the default skill handler wrapper.

    This wrapper handles things like metrics, reporting handler start/stop
    and errors.

    Args:
        handler (callable): method/function to call
        on_error (function): function to call to report error.

    Returns:
        Wrapped callable
    """

    def wrapper(message):
        try:
            if len(signature(handler).parameters) == 0:
                handler()
            else:
                handler(message)
        except Exception as e:
            if on_error:
                on_error(e)

    return wrapper


class EventContainer:
    """Container tracking messagbus handlers.

    This container tracks events added by a skill, allowing unregistering
    all events on shutdown.
    """

    def __init__(self, bus=None):
        self.bus = bus
        self.events = []

    def set_bus(self, bus):
        self.bus = bus

    def add(self, name, handler, once=False):
        """Create event handler for executing intent or other event.

        Args:
            name (string): IntentParser name
            handler (func): Method to call
            once (bool, optional): Event handler will be removed after it has
                                   been run once.
        """

        def once_wrapper(message):
            # Remove registered one-time handler before invoking,
            # allowing them to re-schedule themselves.
            self.remove(name)
            handler(message)

        if handler:
            if once:
                self.bus.once(name, once_wrapper)
                self.events.append((name, once_wrapper))
            else:
                self.bus.on(name, handler)
                self.events.append((name, handler))

            LOG.debug('Added event: {}'.format(name))

    def remove(self, name):
        """Removes an event from bus emitter and events list.

        Args:
            name (string): Name of Intent or Scheduler Event
        Returns:
            bool: True if found and removed, False if not found
        """
        LOG.debug("Removing event {}".format(name))
        removed = False
        for _name, _handler in list(self.events):
            if name == _name:
                try:
                    self.events.remove((_name, _handler))
                except ValueError:
                    LOG.error('Failed to remove event {}'.format(name))
                    pass
                removed = True

        # Because of function wrappers, the emitter doesn't always directly
        # hold the _handler function, it sometimes holds something like
        # 'wrapper(_handler)'.  So a call like:
        #     self.bus.remove(_name, _handler)
        # will not find it, leaving an event handler with that name left behind
        # waiting to fire if it is ever re-installed and triggered.
        # Remove all handlers with the given name, regardless of handler.
        if removed:
            self.bus.remove_all_listeners(name)
        return removed

    def __iter__(self):
        return iter(self.events)

    def clear(self):
        """Unregister all registered handlers and clear the list of registered
        events.
        """
        for e, f in self.events:
            self.bus.remove(e, f)
        self.events = []  # Remove reference to wrappers


class BusService:
    """
    Provide some service over the messagebus for other components

    response = Message("face.recognition.reply")
    service = BusService(response)
    service.listen("face.recognition")

    while True:
        data = do_computation()
        service.update_response(data)  # replaces response.data

    """

    def __init__(self, message, trigger_messages=None, bus=None):
        self.bus = bus or get_mycroft_bus()
        self.response = message
        trigger_messages = trigger_messages or []
        self.events = []
        for message_type in trigger_messages:
            self.listen(message_type)

    def listen(self, message_type, callback=None):
        if callback is None:
            callback = self._respond
        self.bus.on(message_type, callback)
        self.events.append((message_type, callback))

    def update_response(self, data=None):
        if data is not None:
            self.response.data = data

    def _respond(self, message):
        self.bus.emit(message.reply(self.response.type, self.response.data))

    def shutdown(self):
        """ remove all listeners """
        for event, callback in self.events:
            self.bus.remove(event, callback)


class BusFeedProvider:
    """

    Meant to be subclassed


        class ClockService(BusFeedProvider):
            def __init__(self, name="clock_transmitter", bus=None):
                trigger_message  = Message("time.request")
                super().__init__(trigger_message, name, bus)
                self.set_data_gatherer(self.handle_get_time)

            def handle_get_time(self, message):
                self.update({"date": datetime.now()})


        clock_service = ClockService()

    """

    def __init__(self, trigger_message, name=None, bus=None, config=None):
        """
           initialize responder

           args:
                name(str): name identifier for .conf settings
                bus (WebsocketClient): mycroft messagebus websocket
        """
        if not config:
            LOG.warning(f"Expected a dict config and got None. This config"
                        f"fallback behavior will be deprecated in a future "
                        f"release")
            try:
                from ovos_config.config import read_mycroft_config
                config = read_mycroft_config()
            except ImportError:
                LOG.warning("ovos_config not available. Falling back to "
                            "default configuration")
                config = dict()
        self.trigger_message = trigger_message
        self.name = name or self.__class__.__name__
        self.bus = bus or get_mycroft_bus()
        self.callback = None
        self.service = None
        self._daemon = None
        self.config = config.get(self.name, {})

    def update(self, data):
        """
        change the data of the response to be sent when queried
        """
        if self.service is not None:
            self.service.update_response(data)

    def set_data_gatherer(self, callback, default_data=None, daemonic=False, interval=90):
        """
          prepare responder for sending, register answers
        """

        self.bus.remove_all_listeners(self.trigger_message)
        if ".request" in self.trigger_message:
            response_type = self.trigger_message.replace(".request", ".reply")
        else:
            response_type = self.trigger_message + ".reply"

        response = FakeMessage(response_type, default_data)
        self.service = BusService(response, bus=self.bus)
        self.callback = callback
        self.bus.on(self.trigger_message, self._respond)
        if daemonic:
            self._daemon = create_loop(self._data_daemon, interval)

    def _data_daemon(self):
        if self.callback is not None:
            self.callback(self.trigger_message)

    def _respond(self, message):
        """
          gather data and emit to bus
        """
        try:
            if self.callback:
                self.callback(message)
        except Exception as e:
            LOG.error(e)
        self.service.respond(message)

    def shutdown(self):
        self.bus.remove_all_listeners(self.trigger_message)
        if self._daemon:
            self._daemon.join(0)
            self._daemon = None
        if self.service:
            self.service.shutdown()
            self.service = None


class BusQuery:
    """
    retrieve data from some other component over the messagebus at any time

    message = Message("request.msg", {...}, {...})
    query = BusQuery(message)
    response = query.send()
    # do some more stuff
    response = query.send() # reutilize the object

    """

    def __init__(self, message, bus=None):
        self.bus = bus or get_mycroft_bus()
        self._waiting = False
        self.response = FakeMessage(None, None, None)
        self.query = message
        self.valid_response_types = []

    def add_response_type(self, response_type):
        """ listen to a new response_type """
        if response_type not in self.valid_response_types:
            self.valid_response_types.append(response_type)
            self.bus.on(response_type, self._end_wait)

    def _end_wait(self, message):
        self.response = message
        self._waiting = False

    def _wait_response(self, timeout):
        start = time.time()
        elapsed = 0
        self._waiting = True
        while self._waiting and elapsed < timeout:
            elapsed = time.time() - start
            time.sleep(0.1)
        self._waiting = False

    def send(self, response_type=None, timeout=10):
        self.response = FakeMessage(None, None, None)
        if response_type is None:
            response_type = self.query.type + ".reply"
        self.add_response_type(response_type)
        self.bus.emit(self.query)
        self._wait_response(timeout)
        return self.response

    def remove_listeners(self):
        for event in self.valid_response_types:
            self.bus.remove(event, self._end_wait)

    def shutdown(self):
        """ remove all listeners """
        self.remove_listeners()


class BusFeedConsumer:
    """
    this is meant to be subclassed

    class Clock(BusFeedConsumer):
        def __init__(self, name="clock_receiver", timeout=3, bus=None):
            request_message = Message("time.request")
            super().__init__(request_message, name, timeout, bus)

    # blocking
    clock = Clock()
    date = clock.request()["date"]

    # async
    clock = Clock(timeout=0) # non - blocking
    clock.request(daemonic=True, # loop on background
                  frequency=1) # update result every second

    date = clock.result["date"]

    """

    def __init__(self, query_message, name=None, timeout=5, bus=None,
                 config=None):
        self.query_message = query_message
        self.query_message.context["source"] = self.name
        self.name = name or self.__class__.__name__
        self.bus = bus or get_mycroft_bus()
        if not config:
            try:
                from ovos_config.config import read_mycroft_config
                config = read_mycroft_config()
            except ImportError:
                LOG.warning("Config not provided and ovos_config not available")
                config = dict()
        self.config = config.get(self.name, {})
        self.timeout = timeout
        self.query = None
        self.valid_responses = []
        self._daemon = None

    def request(self, response_messages=None, daemonic=False, interval=90):
        """
          prepare query for sending, add several possible kinds of
          response message automatically
                "message_type.reply" ,
                "message_type.response",
                "message_type.result"
        """
        response_messages = response_messages or []

        # generate valid reply message types
        self.valid_responses = response_messages
        if ".request" in self.query_message.type:
            response = self.query_message.type.replace(".request", ".reply")
            if response not in self.valid_responses:
                self.valid_responses.append(response)
            response = self.query_message.type.replace(".request", ".response")
            if response not in self.valid_responses:
                self.valid_responses.append(response)
            response = self.query_message.type.replace(".request", ".result")
            if response not in self.valid_responses:
                self.valid_responses.append(response)
        else:
            response = self.query_message.type + ".reply"
            if response not in self.valid_responses:
                self.valid_responses.append(response)
            response = self.query_message.type + ".response"
            if response not in self.valid_responses:
                self.valid_responses.append(response)
            response = self.query_message.type + ".result"
            if response not in self.valid_responses:
                self.valid_responses.append(response)

        # update message context
        self.query_message.context["valid_responses"] = self.valid_responses

        self._query()
        if daemonic:
            self._daemon = create_loop(self._request_daemon, interval)
        return self.result

    def _request_daemon(self):
        self.query.send(self.valid_responses[0], self.timeout)

    def _query(self):
        self.query = BusQuery(self.query_message)
        for message in self.valid_responses[1:]:
            self.query.add_response_type(message)
        self.query.send(self.valid_responses[0], self.timeout)

    @property
    def result(self):
        return self.query.response.data

    def shutdown(self):
        """ remove all listeners """
        if self._daemon:
            self._daemon.join(0)
            self._daemon = None
        if self.query:
            self.query.shutdown()



from ovos_config.config import read_mycroft_config, update_mycroft_config
from ovos_utils.messagebus import wait_for_reply
from ovos_utils.skills.locations import get_default_skills_directory, get_installed_skill_ids
from ovos_utils.log import LOG


def get_non_properties(obj):
    """Get attributes that are not properties from object.

    Will return members of object class along with bases down to MycroftSkill.

    Args:
        obj: object to scan

    Returns:
        Set of attributes that are not a property.
    """

    def check_class(cls):
        """Find all non-properties in a class."""
        # Current class
        d = cls.__dict__
        np = [k for k in d if not isinstance(d[k], property)]
        # Recurse through base classes excluding MycroftSkill and object
        for b in [b for b in cls.__bases__ if b.__name__ not in ("object", "MycroftSkill")]:
            np += check_class(b)
        return np

    return set(check_class(obj.__class__))


def skills_loaded(bus=None):
    reply = wait_for_reply('mycroft.skills.all_loaded',
                           'mycroft.skills.all_loaded.response',
                           bus=bus)
    if reply:
        return reply.data['status']
    return False


def blacklist_skill(skill, config=None):
    config = config or read_mycroft_config()
    skills_config = config.get("skills", {})
    blacklisted_skills = skills_config.get("blacklisted_skills", [])
    if skill not in blacklisted_skills:
        blacklisted_skills.append(skill)
        conf = {
            "skills": {
                "blacklisted_skills": blacklisted_skills
            }
        }
        update_mycroft_config(conf)
        return True
    return False


def whitelist_skill(skill, config=None):
    config = config or read_mycroft_config()
    skills_config = config.get("skills", {})
    blacklisted_skills = skills_config.get("blacklisted_skills", [])
    if skill in blacklisted_skills:
        blacklisted_skills.pop(skill)
        conf = {
            "skills": {
                "blacklisted_skills": blacklisted_skills
            }
        }
        update_mycroft_config(conf)
        return True
    return False


def make_priority_skill(skill, config=None):
    # TODO: Deprecate in 0.1.0
    LOG.warning("This method is deprecated. Skills are now loaded based on "
                "`runtime_requirements`")
    # config = config or read_mycroft_config()
    # skills_config = config.get("skills", {})
    # priority_skills = skills_config.get("priority_skills", [])
    # if skill not in priority_skills:
    #     priority_skills.append(skill)
    #     conf = {
    #         "skills": {
    #             "priority_skills": priority_skills
    #         }
    #     }
    #     update_mycroft_config(conf)
    #     return True
    return False


def get_skills_folder(config=None):
    LOG.warning("This reference is deprecated, use "
                "`ovos_utils.skills.locations.get_default_skill_dir")
    return get_default_skills_directory(config)


def get_installed_skills(config=None):
    LOG.warning("This reference is deprecated, use "
                "`ovos_utils.skills.locations.get_installed_skill_ids")
    return get_installed_skill_ids(config)



import os
import random
import re
from os.path import join
from pathlib import Path
from typing import Optional

from ovos_utils.bracket_expansion import expand_options
from ovos_utils.file_utils import resolve_resource_file
from ovos_utils.lang import translate_word
from ovos_utils.log import LOG


class MustacheDialogRenderer:
    """A dialog template renderer based on the mustache templating language."""

    def __init__(self):
        self.templates = {}
        self.recent_phrases = []

        # TODO magic numbers are bad!
        self.max_recent_phrases = 3
        # We cycle through lines in .dialog files to keep Mycroft from
        # repeating the same phrase over and over. However, if a .dialog
        # file only contains a few entries, this can cause it to loop.
        #
        # This offset will override max_recent_phrases on very short .dialog
        # files. With the offset at 2, .dialog files with 3 or more lines will
        # be managed to avoid repetition, but .dialog files with only 1 or 2
        # lines will be unaffected. Dialog should never get stuck in a loop.
        self.loop_prevention_offset = 2

    def load_template_file(self, template_name, filename):
        """Load a template by file name into the templates cache.

        Args:
            template_name (str): a unique identifier for a group of templates
            filename (str): a fully qualified filename of a mustache template.
        """
        with open(filename, 'r', encoding='utf8') as f:
            for line in f:
                template_text = line.strip()
                # Skip all lines starting with '#' and all empty lines
                if (not template_text.startswith('#') and
                        template_text != ''):
                    if template_name not in self.templates:
                        self.templates[template_name] = []

                    # convert to standard python format string syntax. From
                    # double (or more) '{' followed by any number of
                    # whitespace followed by actual key followed by any number
                    # of whitespace followed by double (or more) '}'
                    template_text = re.sub(r'\{\{+\s*(.*?)\s*\}\}+', r'{\1}',
                                           template_text)

                    self.templates[template_name].append(template_text)

    def render(self, template_name, context=None, index=None):
        """
        Given a template name, pick a template and render it using the context.
        If no matching template exists use template_name as template.

        Tries not to let Mycroft say exactly the same thing twice in a row.

        Args:
            template_name (str): the name of a template group.
            context (dict): dictionary representing values to be rendered
            index (int): optional, the specific index in the collection of
                templates

        Returns:
            str: the rendered string
        """
        context = context or {}
        if template_name not in self.templates:
            # When not found, return the name itself as the dialog
            # This allows things like render("record.not.found") to either
            # find a translation file "record.not.found.dialog" or return
            # "record not found" literal.
            return template_name.replace('.', ' ')

        # Get the .dialog file's contents, minus any which have been spoken
        # recently.
        template_functions = self.templates.get(template_name)

        if index is None:
            template_functions = ([t for t in template_functions
                                   if t not in self.recent_phrases] or
                                  template_functions)
            line = random.choice(template_functions)
        else:
            line = template_functions[index % len(template_functions)]
        # Replace {key} in line with matching values from context
        line = line.format(**context)
        line = random.choice(expand_options(line))

        # Here's where we keep track of what we've said recently. Remember,
        # this is by line in the .dialog file, not by exact phrase
        self.recent_phrases.append(line)
        if (len(self.recent_phrases) >
                min(self.max_recent_phrases, len(self.templates.get(
                    template_name)) - self.loop_prevention_offset)):
            self.recent_phrases.pop(0)
        return line


def load_dialogs(dialog_dir: str,
                 renderer: Optional[MustacheDialogRenderer] = None) -> \
        MustacheDialogRenderer:
    """
    Load all dialog files within the specified directory.

    Args:
        dialog_dir (str): directory that contains dialog files
        renderer (MustacheDialogRenderer): instance to load files with
    Returns:
        a loaded instance of a dialog renderer
    """
    if renderer is None:
        renderer = MustacheDialogRenderer()

    directory = Path(dialog_dir)
    if not directory.exists() or not directory.is_dir():
        LOG.warning('No dialog files found: {}'.format(dialog_dir))
        return renderer

    for path, _, files in os.walk(str(directory)):
        for f in files:
            if f.endswith('.dialog'):
                renderer.load_template_file(f.replace('.dialog', ''),
                                            join(path, f))
    return renderer


def get_dialog(phrase: str, lang: str = None,
               context: Optional[dict] = None) -> str:
    """
    Looks up a resource file for the given phrase in the specified language.

    If no file is found, the requested phrase is returned as the string. This
    will use the default language for translations.

    Args:
        phrase (str): resource phrase to retrieve/translate
        lang (str): the language to use
        context (dict): values to be inserted into the string

    Returns:
        str: a randomized and/or translated version of the phrase
    """

    if not lang:
        LOG.warning(f"Expected a string lang and got None. This config"
                    f"fallback behavior will be deprecated in a future release")
        try:
            from ovos_config.config import read_mycroft_config
            conf = read_mycroft_config()
            lang = conf.get('lang')
        except ImportError:
            LOG.warning("Config not provided and ovos_config not available")
            lang = "en-us"
        except FileNotFoundError:
            LOG.warning("Configuration file not found, default lang to 'en-us'")
            lang = "en-us"

    filename = join('text', lang.lower(), phrase + '.dialog')
    template = resolve_resource_file(filename)
    if not template:
        LOG.debug('Resource file not found: {}'.format(filename))
        return phrase

    stache = MustacheDialogRenderer()
    stache.load_template_file('template', template)
    if not context:
        context = {}
    return stache.render('template', context)


def join_list(items: list, connector: str, sep: Optional[str] = None,
              lang: Optional[str] = '') -> str:
    """
    Join a list into a phrase using the given connector word
    Examples:
        join_list([1,2,3], "and") ->  "1, 2 and 3"
        join_list([1,2,3], "and", ";") ->  "1; 2 and 3"
    Args:
        items (array): items to be joined
        connector (str): connecting word (resource name), like "and" or "or"
        sep (str, optional): separator character, default = ","
        lang (str, optional): an optional BCP-47 language code, if omitted
                              the default language will be used.
    Returns:
        str: the connected list phrase
    """

    if not items:
        return ""
    if len(items) == 1:
        return str(items[0])

    if not sep:
        sep = ", "
    else:
        sep += " "
    return (sep.join(str(item) for item in items[:-1]) +
            " " + translate_word(connector, lang) +
            " " + items[-1])



import inspect
import os
import re
import shutil
import subprocess
import sys
import sysconfig
from enum import Enum
from os.path import expanduser, exists, join

from ovos_utils.log import LOG


# TODO: Deprecate MycroftRootLocations in 0.1.0
class MycroftRootLocations(str, Enum):
    PICROFT = "/home/pi/mycroft-core"
    BIGSCREEN = "/home/mycroft/mycroft-core"
    OVOS = "/usr/lib/python3.9/site-packages"
    OLD_MARK1 = "/opt/venvs/mycroft-core/lib/python3.4/site-packages"
    MARK1 = "/opt/venvs/mycroft-core/lib/python3.7/site-packages"
    MARK2 = "/opt/mycroft"
    HOME = expanduser("~/mycroft-core")  # git clones


_USER_DEFINED_ROOT = None


def is_running_from_module(module_name):
    # Stack:
    # [0] - _log()
    # [1] - debug(), info(), warning(), or error()
    # [2] - caller
    stack = inspect.stack()

    # Record:
    # [0] - frame object
    # [1] - filename
    # [2] - line number
    # [3] - function
    # ...
    for record in stack[2:]:
        mod = inspect.getmodule(record[0])
        name = mod.__name__ if mod else ''
        # module name in file path of caller
        # or import name matches module name
        if f"/{module_name}/" in record[1] or \
                name.startswith(module_name.replace("-", "_").replace(" ", "_")):
            return True
    return False


# system utils
def ntp_sync():
    # Force the system clock to synchronize with internet time servers
    subprocess.call('service ntp stop', shell=True)
    subprocess.call('ntpd -gq', shell=True)
    subprocess.call('service ntp start', shell=True)


def system_shutdown(sudo=True):
    """
    Turn the system completely off (with no option to inhibit it)
    @param sudo: use sudo when calling systemctl
    """
    cmd = 'systemctl poweroff -i'
    if sudo:
        cmd = f'sudo {cmd}'
    subprocess.call(cmd, shell=True)


def system_reboot(sudo=True):
    """
    Shut down and restart the system
    @param sudo: use sudo when calling systemctl
    """
    cmd = 'systemctl reboot -i'
    if sudo:
        cmd = f'sudo {cmd}'
    subprocess.call(cmd, shell=True)


def ssh_enable(sudo=True, user=False):
    """
    Permanently allow SSH access
    @param sudo: use sudo when calling systemctl
    @param user: pass --user flag when calling systemctl
    """
    enable_service("ssh.service", sudo=sudo, user=user)


def ssh_disable(sudo=True, user=False):
    """
    Permanently block SSH access from the outside
    @param sudo: use sudo when calling systemctl
    @param user: pass --user flag when calling systemctl
    """
    disable_service("ssh.service", sudo=sudo, user=user)


def restart_mycroft_service(sudo=True, user=False):
    """
    Restarts the `mycroft.service` systemd service
    @param sudo: use sudo when calling systemctl
    @param user: pass --user flag when calling systemctl
    """
    restart_service("mycroft.service", sudo=sudo, user=user)


def restart_service(service_name, sudo=True, user=False):
    """
    Restarts a systemd service using systemctl
    @param service_name: name of service to restart
    @param sudo: use sudo when calling systemctl
    @param user: pass --user flag when calling systemctl
    """
    cmd = f'systemctl restart {service_name}'
    if user:
        cmd = f"{cmd} --user"
    elif sudo:
        cmd = f"sudo {cmd}"
    subprocess.call(cmd, shell=True)


def enable_service(service_name, sudo=False, user=False):
    """
    Enables and Starts a systemd service using systemctl
    @param service_name: name of service to Enable and Start
    @param sudo: use sudo when calling systemctl
    @param user: pass --user flag when calling systemctl
    """
    enable_command = f"systemctl enable {service_name}"
    start_command = f"systemctl start {service_name}"
    if user:
        enable_command = f"{enable_command} --user"
        start_command = f"{start_command} --user"
    elif sudo:
        enable_command = f"sudo {enable_command}"
        start_command = f"sudo {start_command}"
    subprocess.call(enable_command, shell=True)
    subprocess.call(start_command, shell=True)


def disable_service(service_name, sudo=False, user=False):
    """
    Disables and Stops a systemd service using systemctl
    @param service_name: name of service to Disable and Stop
    @param sudo: use sudo when calling systemctl
    @param user: pass --user flag when calling systemctl
    """
    disable_command = f"systemctl disable {service_name}"
    stop_command = f"systemctl stop {service_name}"
    if user:
        disable_command = f"{disable_command} --user"
        stop_command = f"{stop_command} --user"
    elif sudo:
        disable_command = f"sudo {disable_command}"
        stop_command = f"sudo {stop_command}"
    subprocess.call(stop_command, shell=True)
    subprocess.call(disable_command, shell=True)


def check_service_active(service_name, sudo=False, user=False) -> bool:
    """
    Checks if a systemd service is active using systemctl
    @param service_name: name of service to check
    @param user: pass --user flag when calling systemctl
    @param sudo: use sudo when calling systemctl
    @return: True if the service is active, else False
    """
    status_command = f"systemctl is-active --quiet {service_name}"
    if user:
        status_command = f"{status_command} --user"
    elif sudo:
        status_command = f"sudo {status_command}"
    state = subprocess.run(status_command, shell=True).returncode
    return state == 0


# platform fingerprinting
def set_root_path(path):
    global _USER_DEFINED_ROOT
    _USER_DEFINED_ROOT = path
    LOG.info(f"mycroft root set to {path}")


def find_root_from_sys_path():
    """Find mycroft root folder from sys.path, eg. venv site-packages."""
    for p in [path for path in sys.path if path != '']:
        if exists(join(p, 'mycroft', 'configuration', 'mycroft.conf')):
            return p
    else:
        return None


def find_root_from_sitepackages():
    """Find root from system or venv's sitepackages."""
    site = sysconfig.get_paths()['platlib']
    if exists(join(site, 'mycroft', 'configuration', 'mycroft.conf')):
        return site
    else:
        return None


def search_mycroft_core_location():
    """Check python path (.venv), system packages and finally known mycroft
    locations."""
    # downstream wants to override the root location
    if _USER_DEFINED_ROOT:
        return _USER_DEFINED_ROOT
    # if we are in a .venv that should take precedence over everything else
    if find_root_from_sitepackages():
        return find_root_from_sitepackages()
    # if there is a system wide install that should take precedence over
    # hardcoded locations
    elif find_root_from_sys_path():
        return find_root_from_sys_path()
    # finally look at default locations
    for p in MycroftRootLocations:
        if os.path.isdir(p):
            return p
    return None


def get_desktop_environment():
    # From http://stackoverflow.com/questions/2035657/what-is-my-current-desktop-environment
    # and http://ubuntuforums.org/showthread.php?t=652320
    # and http://ubuntuforums.org/showthread.php?t=652320
    # and http://ubuntuforums.org/showthread.php?t=1139057
    if sys.platform in ["win32", "cygwin"]:
        return "windows"
    elif sys.platform == "darwin":
        return "mac"
    else:  # Most likely either a POSIX system or something not much common
        desktop_session = os.environ.get("DESKTOP_SESSION")
        if desktop_session is not None:  # easier to match if we doesn't have  to deal with character cases
            desktop_session = desktop_session.lower()
            if desktop_session in ["gnome", "unity", "cinnamon", "mate",
                                   "xfce4", "lxde", "fluxbox",
                                   "blackbox", "openbox", "icewm", "jwm",
                                   "afterstep", "trinity", "kde"]:
                return desktop_session
            # Special cases
            # Canonical sets $DESKTOP_SESSION to Lubuntu rather than LXDE if using LXDE.
            # There is no guarantee that they will not do the same with the other desktop environments.
            elif "xfce" in desktop_session or desktop_session.startswith(
                    "xubuntu"):
                return "xfce4"
            elif desktop_session.startswith("ubuntu"):
                return "unity"
            elif desktop_session.startswith("lubuntu"):
                return "lxde"
            elif desktop_session.startswith("kubuntu"):
                return "kde"
            elif desktop_session.startswith("razor"):  # e.g. razorkwin
                return "razor-qt"
            elif desktop_session.startswith("wmaker"):  # e.g. wmaker-common
                return "windowmaker"
        if os.environ.get('KDE_FULL_SESSION') == 'true':
            return "kde"
        elif os.environ.get('GNOME_DESKTOP_SESSION_ID'):
            if not "deprecated" in os.environ.get('GNOME_DESKTOP_SESSION_ID'):
                return "gnome2"
        # From http://ubuntuforums.org/showthread.php?t=652320
        elif is_process_running("xfce-mcs-manage"):
            return "xfce4"
        elif is_process_running("ksmserver"):
            return "kde"
        elif is_process_running("icewm"):
            return "icewm"
        elif is_process_running("fluxbox"):
            return "fluxbox"
        elif is_process_running("jwm"):
            return "jwm"
    return "unknown"


def is_process_running(process):
    # From http://www.bloggerpolis.com/2011/05/how-to-check-if-a-process-is-running-using-python/
    # and http://richarddingwall.name/2009/06/18/windows-equivalents-of-ps-and-kill-commands/
    try:  # Linux/Unix
        s = subprocess.Popen(["ps", "axw"], stdout=subprocess.PIPE)
    except:  # Windows
        s = subprocess.Popen(["tasklist", "/v"], stdout=subprocess.PIPE)
    for x in s.stdout:
        if re.search(process, x.decode("utf-8")):
            return True
    return False


def find_executable(executable):
    return shutil.which(executable)


def is_installed(executable):
    return bool(find_executable(executable))


def has_screen():
    have_display = "DISPLAY" in os.environ
    # X server not running
    if not have_display:
        # raspberry pi specific check
        try:
            have_display = b"device_name=" in subprocess.check_output("tvservice -n 2>&1", shell=True)
        except Exception as e:
            pass
        
    # fallback check using matplotlib if available
    # seems to be foolproof and OS agnostic 
    # but do not want to drag the dependency
    if not have_display:
        try:
            import matplotlib.pyplot as plt
            try:
                plt.figure()
                have_display = True
            except:
                have_display = False
        except ImportError:
            pass
    return have_display


def module_property(func):
    """
    Decorator to turn module functions into properties.
    Function names must be prefixed with an underscore.
    :param func: function to decorate
    """
    import sys
    module = sys.modules[func.__module__]

    def fallback_getattr(name):
        raise AttributeError(
            f"module '{module.__name__}' has no attribute '{name}'")

    default_getattr = getattr(module, '__getattr__', fallback_getattr)

    def patched_getattr(name):
        if f'_{name}' == func.__name__:
            return func()
        return default_getattr(name)

    module.__getattr__ = patched_getattr
    return func




# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime
import re
from functools import lru_cache, wraps
from os.path import isdir, join
from threading import Thread, Event
from time import monotonic_ns
from time import sleep

import kthread

# TODO: Deprecate below imports
from ovos_utils.file_utils import resolve_ovos_resource_file, resolve_resource_file
from ovos_utils.network_utils import get_ip, get_external_ip, is_connected_dns, is_connected_http, is_connected
from ovos_utils.log import LOG


class classproperty(property):
    """Decorator for a Class-level property.
    Credit to Denis Rhyzhkov on Stackoverflow: https://stackoverflow.com/a/13624858/1280629"""
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


def ensure_mycroft_import():
    # TODO: Deprecate in 0.1.0
    LOG.warning("This method is deprecated. Anything depending on `mycroft`"
                "should install `ovos-core` as a dependency")
    try:
        import mycroft
    except ImportError:
        import sys
        from ovos_utils import get_mycroft_root
        MYCROFT_ROOT_PATH = get_mycroft_root()
        if MYCROFT_ROOT_PATH is not None:
            sys.path.append(MYCROFT_ROOT_PATH)
        else:
            raise


def get_mycroft_root():
    # TODO: Deprecate in 0.1.0
    LOG.warning("This method is deprecated. Code should import from the current"
                "namespace; other system paths are irrelevant.")
    paths = [
        "/opt/venvs/mycroft-core/lib/python3.7/site-packages/",  # mark1/2
        "/opt/venvs/mycroft-core/lib/python3.4/site-packages/ ",  # old mark1 installs
        "/home/pi/mycroft-core"  # picroft
    ]
    for p in paths:
        if isdir(join(p, "mycroft")):
            return p
    return None


def timed_lru_cache(
        _func=None, *, seconds: int = 7000, maxsize: int = 128, typed: bool = False
):
    """ Extension over existing lru_cache with timeout

    taken from: https://blog.soumendrak.com/cache-heavy-computation-functions-with-a-timeout-value

    :param seconds: timeout value
    :param maxsize: maximum size of the cache
    :param typed: whether different keys for different types of cache keys
    """

    def wrapper_cache(f):
        # create a function wrapped with traditional lru_cache
        f = lru_cache(maxsize=maxsize, typed=typed)(f)
        # convert seconds to nanoseconds to set the expiry time in nanoseconds
        f.delta = seconds * 10 ** 9
        f.expiration = monotonic_ns() + f.delta

        @wraps(f)  # wraps is used to access the decorated function attributes
        def wrapped_f(*args, **kwargs):
            if monotonic_ns() >= f.expiration:
                # if the current cache expired of the decorated function then
                # clear cache for that function and set a new cache value with new expiration time
                f.cache_clear()
                f.expiration = monotonic_ns() + f.delta
            return f(*args, **kwargs)

        wrapped_f.cache_info = f.cache_info
        wrapped_f.cache_clear = f.cache_clear
        return wrapped_f

    # To allow decorator to be used without arguments
    if _func is None:
        return wrapper_cache
    else:
        return wrapper_cache(_func)


def create_killable_daemon(target, args=(), kwargs=None, autostart=True):
    """Helper to quickly create and start a thread with daemon = True"""
    t = kthread.KThread(target=target, args=args, kwargs=kwargs)
    t.daemon = True
    if autostart:
        t.start()
    return t


def create_daemon(target, args=(), kwargs=None, autostart=True):
    """Helper to quickly create and start a thread with daemon = True"""
    t = Thread(target=target, args=args, kwargs=kwargs)
    t.daemon = True
    if autostart:
        t.start()
    return t


def create_loop(target, interval, args=(), kwargs=None):
    """
    Helper to quickly create and start a thread with daemon = True
    and repeat it every interval seconds
    """

    def loop(*args, **kwargs):
        try:
            while True:
                target(*args, **kwargs)
                sleep(interval)
        except KeyboardInterrupt:
            return

    return create_daemon(loop, args, kwargs)


def wait_for_exit_signal():
    """Blocks until KeyboardInterrupt is received"""
    try:
        Event().wait()
    except KeyboardInterrupt:
        pass


def get_handler_name(handler):
    """Name (including class if available) of handler function.

    Arguments:
        handler (function): Function to be named

    Returns:
        string: handler name as string
    """
    if '__self__' in dir(handler) and 'name' in dir(handler.__self__):
        return handler.__self__.name + '.' + handler.__name__
    else:
        return handler.__name__


def camel_case_split(identifier: str) -> str:
    """Split camel case string"""
    regex = '.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)'
    matches = re.finditer(regex, identifier)
    return ' '.join([m.group(0) for m in matches])


def rotate_list(l, n=1):
    return l[n:] + l[:n]


def flatten_list(some_list, tuples=True):
    _flatten = lambda l: [item for sublist in l for item in sublist]
    if tuples:
        while any(isinstance(x, list) or isinstance(x, tuple)
                  for x in some_list):
            some_list = _flatten(some_list)
    else:
        while any(isinstance(x, list) for x in some_list):
            some_list = _flatten(some_list)
    return some_list


def datestr2ts(datestr):
    y = int(datestr[:4])
    m = int(datestr[4:6])
    d = int(datestr[-2:])
    dt = datetime.datetime(y, m, d)
    return dt.timestamp()



import os
import subprocess
import time

from copy import deepcopy
from distutils.spawn import find_executable
from ovos_utils.file_utils import resolve_resource_file
from ovos_utils.log import LOG
from ovos_utils.signal import check_for_signal

try:
    from ovos_config.config import read_mycroft_config
except ImportError:
    LOG.warning("Config not provided and ovos_config not available")

    def read_mycroft_config():
        return dict()

# Create a custom environment to use that can be ducked by a phone role.
# This is kept separate from the normal os.environ to ensure that the TTS
# role isn't affected and that any thirdparty software launched through
# a mycroft process can select if they wish to honor this.
_ENVIRONMENT = deepcopy(os.environ)
_ENVIRONMENT['PULSE_PROP'] = 'media.role=music'


def _get_pulse_environment(config):
    """Return environment for pulse audio depeding on ducking config."""
    tts_config = config.get('tts', {})
    if tts_config and tts_config.get('pulse_duck'):
        return _ENVIRONMENT
    else:
        return os.environ


def play_acknowledge_sound():
    """Acknowledge a successful request.

    This method plays a sound to acknowledge a request that does not
    require a verbal response. This is intended to provide simple feedback
    to the user that their request was handled successfully.
    """
    audio_file = resolve_resource_file(
        read_mycroft_config().get('sounds', {}).get('acknowledge'))

    if not audio_file:
        LOG.warning("Could not find 'acknowledge' audio file!")
        return

    process = play_audio(audio_file)
    if not process:
        LOG.warning("Unable to play 'acknowledge' audio file!")
    return process


def play_listening_sound():
    """Audibly indicate speech recording started."""
    audio_file = resolve_resource_file(
        read_mycroft_config().get('sounds', {}).get('start_listening'))

    if not audio_file:
        LOG.warning("Could not find 'start_listening' audio file!")
        return

    process = play_audio(audio_file)
    if not process:
        LOG.warning("Unable to play 'start_listening' audio file!")
    return process


def play_end_listening_sound():
    """Audibly indicate speech recording is no longer happening."""
    audio_file = resolve_resource_file(
        read_mycroft_config().get('sounds', {}).get('end_listening'))

    if not audio_file:
        LOG.debug("Could not find 'end_listening' audio file!")
        return

    process = play_audio(audio_file)
    if not process:
        LOG.warning("Unable to play 'end_listening' audio file!")
    return process


def play_error_sound():
    """Audibly indicate a failed request.

    This method plays a error sound to signal an error that does not
    require a verbal response. This is intended to provide simple feedback
    to the user that their request was NOT handled successfully.
    """
    audio_file = resolve_resource_file(
        read_mycroft_config().get('sounds', {}).get('error'))

    if not audio_file:
        LOG.warning("Could not find 'error' audio file!")
        return

    process = play_audio(audio_file)
    if not process:
        LOG.warning("Unable to play 'error' audio file!")
    return process


def _find_player(uri):
    _, ext = os.path.splitext(uri)

    # scan installed executables that can handle playback
    sox_play = find_executable("play")
    pulse_play = find_executable("paplay")
    alsa_play = find_executable("aplay")
    mpg123_play = find_executable("mpg123")
    ogg123_play = find_executable("ogg123")

    player = None
    # sox should handle almost every format, but fails in some urls
    if sox_play:
        player = sox_play + f" --type {ext} %1"
    # determine best available player
    else:
        if "ogg" in ext and ogg123_play:
            player = ogg123_play + " -q %1"
        # wav file
        if 'wav' in ext:
            if pulse_play:
                player = pulse_play + " %1"
            elif alsa_play:
                player = alsa_play + " %1"
        # guess mp3
        elif mpg123_play:
            player = mpg123_play + " %1"
    return player


def play_audio(uri, play_cmd=None, environment=None):
    """Play an audio file.

    This wraps the other play_* functions, choosing the correct one based on
    the file extension. The function will return directly and play the file
    in the background.

    Args:
        uri:    uri to play
        environment (dict): optional environment for the subprocess call

    Returns: subprocess.Popen object. None if the format is not supported or
             an error occurs playing the file.
    """
    config = read_mycroft_config()
    environment = environment or _get_pulse_environment(config)

    # NOTE: some urls like youtube streams will cause extension detection to fail
    # let's handle it explicitly
    uri = uri.split("?")[0]
    # Replace file:// uri's with normal paths
    uri = uri.replace('file://', '')

    _, ext = os.path.splitext(uri)

    if not play_cmd:
        if "ogg" in ext:
            play_cmd = config.get("play_ogg_cmdline")
        elif "wav" in ext:
            play_cmd = config.get("play_wav_cmdline")
        elif "mp3" in ext:
            play_cmd = config.get("play_mp3_cmdline")

    if not play_cmd:
        play_cmd = _find_player(uri)

    if not play_cmd:
        LOG.error(f"Failed to play: No playback functionality available")
        return None

    play_cmd = play_cmd.split(" ")

    for index, cmd in enumerate(play_cmd):
        if cmd == "%1":
            play_cmd[index] = uri

    try:
        return subprocess.Popen(play_cmd, env=environment)
    except Exception as e:
        LOG.error(f"Failed to play: {play_cmd}")
        LOG.exception(e)
        return None


def play_wav(uri, play_cmd=None, environment=None):
    """ Play a wav-file.

        Returns: subprocess.Popen object
    """
    config = read_mycroft_config()
    environment = environment or _get_pulse_environment(config)
    play_cmd = play_cmd or config.get("play_wav_cmdline") or "paplay %1"
    play_wav_cmd = str(play_cmd).split(" ")
    for index, cmd in enumerate(play_wav_cmd):
        if cmd == "%1":
            play_wav_cmd[index] = uri
    try:
        return subprocess.Popen(play_wav_cmd, env=environment)
    except Exception as e:
        LOG.error("Failed to launch WAV: {}".format(play_wav_cmd))
        LOG.debug("Error: {}".format(repr(e)), exc_info=True)
        return None


def play_mp3(uri, play_cmd=None, environment=None):
    """ Play a mp3-file.

        Returns: subprocess.Popen object
    """
    config = read_mycroft_config()
    environment = environment or _get_pulse_environment(config)
    play_cmd = play_cmd or config.get("play_mp3_cmdline") or "mpg123 %1"
    play_mp3_cmd = str(play_cmd).split(" ")
    for index, cmd in enumerate(play_mp3_cmd):
        if cmd == "%1":
            play_mp3_cmd[index] = uri
    try:
        return subprocess.Popen(play_mp3_cmd, env=environment)
    except Exception as e:
        LOG.error("Failed to launch MP3: {}".format(play_mp3_cmd))
        LOG.debug("Error: {}".format(repr(e)), exc_info=True)
        return None


def play_ogg(uri, play_cmd=None, environment=None):
    """ Play a ogg-file.

        Returns: subprocess.Popen object
    """
    config = read_mycroft_config()
    environment = environment or _get_pulse_environment(config)
    play_cmd = play_cmd or config.get("play_ogg_cmdline") or "ogg123 -q %1"
    play_ogg_cmd = str(play_cmd).split(" ")
    for index, cmd in enumerate(play_ogg_cmd):
        if cmd == "%1":
            play_ogg_cmd[index] = uri
    try:
        return subprocess.Popen(play_ogg_cmd, env=environment)
    except Exception as e:
        LOG.error("Failed to launch OGG: {}".format(play_ogg_cmd))
        LOG.debug("Error: {}".format(repr(e)), exc_info=True)
        return None


def record(file_path, duration, rate, channels):
    """Simple function to record from the default mic.

    The recording is done in the background by the arecord commandline
    application.

    Args:
        file_path: where to store the recorded data
        duration: how long to record
        rate: sample rate
        channels: number of channels

    Returns:
        process for performing the recording.
    """
    command = ['arecord', '-r', str(rate), '-c', str(channels)]
    command += ['-d', str(duration)] if duration > 0 else []
    command += [file_path]
    return subprocess.Popen(command)


def is_speaking():
    """Determine if Text to Speech is occurring

    Returns:
        bool: True while still speaking
    """
    return check_for_signal("isSpeaking", -1)


def wait_while_speaking():
    """Pause as long as Text to Speech is still happening

    Pause while Text to Speech is still happening.  This always pauses
    briefly to ensure that any preceeding request to speak has time to
    begin.
    """
    time.sleep(0.3)  # Wait briefly in for any queued speech to begin
    while is_speaking():
        time.sleep(0.1)



from ovos_utils.system import MycroftRootLocations
from ovos_utils.fingerprinting import detect_platform, MycroftPlatform
from enum import Enum
from os.path import exists
from typing import Optional
from ovos_utils.log import LOG


class MycroftEnclosures(str, Enum):
    # TODO: Deprecate in 0.1.0
    PICROFT = "picroft"
    BIGSCREEN = "kde"
    OVOS = "OpenVoiceOS"
    OLD_MARK1 = "mycroft_mark_1(old)"
    MARK1 = "mycroft_mark_1"
    MARK2 = "mycroft_mark_2"
    HOLMESV = "HolmesV"
    OLD_HOLMES = "mycroft-lib"
    GENERIC = "generic"
    OTHER = "unknown"


def enclosure2rootdir(enclosure: MycroftEnclosures = None) -> Optional[str]:
    """
    Find the default installed core location for a specific platform.
    @param enclosure: MycroftEnclosures object to get root path for
    @return: string default root path
    """
    # TODO: Deprecate in 0.1.0
    LOG.warning("This method is deprecated. Code should import from the current"
                "namespace; other system paths are irrelevant.")
    enclosure = enclosure or detect_enclosure()
    if enclosure == MycroftEnclosures.OLD_MARK1:
        return MycroftRootLocations.OLD_MARK1
    elif enclosure == MycroftEnclosures.MARK1:
        return MycroftRootLocations.MARK1
    elif enclosure == MycroftEnclosures.MARK2:
        return MycroftRootLocations.MARK2
    elif enclosure == MycroftEnclosures.PICROFT:
        return MycroftPlatform.PICROFT
    elif enclosure == MycroftEnclosures.OVOS:
        return MycroftPlatform.OVOS
    elif enclosure == MycroftEnclosures.BIGSCREEN:
        return MycroftPlatform.BIGSCREEN
    return None


def detect_enclosure() -> MycroftEnclosures:
    """
    Determine which enclosure is present on this file system.
    @return: MycroftEnclosures object detected
    """
    # TODO: Deprecate in 0.1.0
    LOG.warning("This method is deprecated. Platform-specific code should"
                "use ovos_utils.fingerprinting.detect_platform directly")
    platform = detect_platform()
    if platform == MycroftPlatform.MARK1:
        if exists(MycroftRootLocations.OLD_MARK1):
            return MycroftEnclosures.OLD_MARK1
        return MycroftEnclosures.MARK1
    elif platform == MycroftPlatform.MARK2:
        return MycroftEnclosures.MARK2
    elif platform == MycroftPlatform.PICROFT:
        return MycroftEnclosures.PICROFT
    elif platform == MycroftPlatform.OVOS:
        return MycroftEnclosures.OVOS
    elif platform == MycroftPlatform.BIGSCREEN:
        return MycroftEnclosures.BIGSCREEN
    elif platform == MycroftPlatform.HOLMESV:
        return MycroftEnclosures.HOLMESV
    elif platform == MycroftPlatform.OLD_HOLMES:
        return MycroftEnclosures.OLD_HOLMES

    return MycroftEnclosures.OTHER



from os import listdir
from os.path import isdir, join

from ovos_utils.file_utils import resolve_resource_file


def get_language_dir(base_path, lang="en-us"):
    """ checks for all language variations and returns best path """
    lang_path = join(base_path, lang)
    # base_path/en-us
    if isdir(lang_path):
        return lang_path
    if "-" in lang:
        main = lang.split("-")[0]
        # base_path/en
        general_lang_path = join(base_path, main)
        if isdir(general_lang_path):
            return general_lang_path
    else:
        main = lang
    # base_path/en-uk, base_path/en-au...
    if isdir(base_path):
        candidates = [join(base_path, f)
                      for f in listdir(base_path) if f.startswith(main)]
        paths = [p for p in candidates if isdir(p)]
        # TODO how to choose best local dialect?
        if len(paths):
            return paths[0]
    return join(base_path, lang)


def translate_word(name, lang='en-us'):
    """ Helper to get word translations
    Args:
        name (str): Word name. Returned as the default value if not translated
        lang (str, optional): an optional BCP-47 language code, if omitted
                              the default language will be used.
    Returns:
        str: translated version of resource name
    """
    filename = resolve_resource_file(join("text", lang, name + ".word"))
    if filename:
        # open the file
        try:
            with open(filename, 'r', encoding='utf8') as f:
                for line in f:
                    word = line.strip()
                    if word.startswith("#"):
                        continue  # skip comment lines
                    return word
        except Exception:
            pass
    return name  # use resource name as the word
